# Form implementation generated from reading ui file '.\raster_scan_base_v2.ui'
#
# Created by: PyQt6 UI code generator 6.9.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_SimpleXYScanWindow(object):
    def setupUi(self, SimpleXYScanWindow):
        SimpleXYScanWindow.setObjectName("SimpleXYScanWindow")
        SimpleXYScanWindow.resize(692, 370)
        SimpleXYScanWindow.setAcceptDrops(False)
        self.centralwidget = QtWidgets.QWidget(parent=SimpleXYScanWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout_2.setVerticalSpacing(3)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.settings_horizontalLayout = QtWidgets.QHBoxLayout()
        self.settings_horizontalLayout.setObjectName("settings_horizontalLayout")
        self.control_groupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.MinimumExpanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.control_groupBox.sizePolicy().hasHeightForWidth()
        )
        self.control_groupBox.setSizePolicy(sizePolicy)
        self.control_groupBox.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.control_groupBox.setObjectName("control_groupBox")
        self.gridLayout = QtWidgets.QGridLayout(self.control_groupBox)
        self.gridLayout.setObjectName("gridLayout")
        self.progressBar = QtWidgets.QProgressBar(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progressBar.sizePolicy().hasHeightForWidth())
        self.progressBar.setSizePolicy(sizePolicy)
        self.progressBar.setProperty("value", 24)
        self.progressBar.setObjectName("progressBar")
        self.gridLayout.addWidget(self.progressBar, 2, 0, 1, 4)
        self.x_label = QtWidgets.QLabel(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.x_label.sizePolicy().hasHeightForWidth())
        self.x_label.setSizePolicy(sizePolicy)
        self.x_label.setObjectName("x_label")
        self.gridLayout.addWidget(self.x_label, 1, 4, 1, 2)
        self.dv_label_2 = QtWidgets.QLabel(parent=self.control_groupBox)
        self.dv_label_2.setObjectName("dv_label_2")
        self.gridLayout.addWidget(self.dv_label_2, 6, 4, 1, 1)
        self.h0_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.h0_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.h0_doubleSpinBox.setSizePolicy(sizePolicy)
        self.h0_doubleSpinBox.setObjectName("h0_doubleSpinBox")
        self.gridLayout.addWidget(self.h0_doubleSpinBox, 4, 2, 1, 1)
        self.Nv_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        self.Nv_doubleSpinBox.setMinimumSize(QtCore.QSize(0, 0))
        self.Nv_doubleSpinBox.setDecimals(0)
        self.Nv_doubleSpinBox.setObjectName("Nv_doubleSpinBox")
        self.gridLayout.addWidget(self.Nv_doubleSpinBox, 6, 7, 1, 1)
        self.dv_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        self.dv_doubleSpinBox.setObjectName("dv_doubleSpinBox")
        self.gridLayout.addWidget(self.dv_doubleSpinBox, 6, 5, 1, 1)
        self.line = QtWidgets.QFrame(parent=self.control_groupBox)
        self.line.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 3, 0, 1, 8)
        self.y_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.y_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.y_doubleSpinBox.setSizePolicy(sizePolicy)
        self.y_doubleSpinBox.setObjectName("y_doubleSpinBox")
        self.gridLayout.addWidget(self.y_doubleSpinBox, 2, 6, 1, 2)
        self.h_label = QtWidgets.QLabel(parent=self.control_groupBox)
        self.h_label.setObjectName("h_label")
        self.gridLayout.addWidget(self.h_label, 4, 0, 1, 1)
        self.Nh_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.Nh_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.Nh_doubleSpinBox.setSizePolicy(sizePolicy)
        self.Nh_doubleSpinBox.setMinimumSize(QtCore.QSize(50, 0))
        self.Nh_doubleSpinBox.setFrame(True)
        self.Nh_doubleSpinBox.setDecimals(0)
        self.Nh_doubleSpinBox.setMaximum(1000.0)
        self.Nh_doubleSpinBox.setProperty("value", 1000.0)
        self.Nh_doubleSpinBox.setObjectName("Nh_doubleSpinBox")
        self.gridLayout.addWidget(self.Nh_doubleSpinBox, 4, 7, 1, 1)
        self.v_label_2 = QtWidgets.QLabel(parent=self.control_groupBox)
        self.v_label_2.setObjectName("v_label_2")
        self.gridLayout.addWidget(self.v_label_2, 6, 0, 1, 1)
        self.v_axis_comboBox = QtWidgets.QComboBox(parent=self.control_groupBox)
        self.v_axis_comboBox.setObjectName("v_axis_comboBox")
        self.gridLayout.addWidget(self.v_axis_comboBox, 6, 1, 1, 1)
        self.start_pushButton = QtWidgets.QPushButton(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.start_pushButton.sizePolicy().hasHeightForWidth()
        )
        self.start_pushButton.setSizePolicy(sizePolicy)
        self.start_pushButton.setObjectName("start_pushButton")
        self.gridLayout.addWidget(self.start_pushButton, 0, 4, 1, 4)
        self.v1_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        self.v1_doubleSpinBox.setObjectName("v1_doubleSpinBox")
        self.gridLayout.addWidget(self.v1_doubleSpinBox, 6, 3, 1, 1)
        self.h1_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.h1_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.h1_doubleSpinBox.setSizePolicy(sizePolicy)
        self.h1_doubleSpinBox.setObjectName("h1_doubleSpinBox")
        self.gridLayout.addWidget(self.h1_doubleSpinBox, 4, 3, 1, 1)
        self.dv_label = QtWidgets.QLabel(parent=self.control_groupBox)
        self.dv_label.setObjectName("dv_label")
        self.gridLayout.addWidget(self.dv_label, 4, 4, 1, 1)
        self.y_label = QtWidgets.QLabel(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.y_label.sizePolicy().hasHeightForWidth())
        self.y_label.setSizePolicy(sizePolicy)
        self.y_label.setObjectName("y_label")
        self.gridLayout.addWidget(self.y_label, 1, 6, 1, 2)
        self.dh_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.dh_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.dh_doubleSpinBox.setSizePolicy(sizePolicy)
        self.dh_doubleSpinBox.setObjectName("dh_doubleSpinBox")
        self.gridLayout.addWidget(self.dh_doubleSpinBox, 4, 5, 1, 1)
        self.label_2 = QtWidgets.QLabel(parent=self.control_groupBox)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 4, 6, 1, 1)
        self.x_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.x_doubleSpinBox.sizePolicy().hasHeightForWidth()
        )
        self.x_doubleSpinBox.setSizePolicy(sizePolicy)
        self.x_doubleSpinBox.setObjectName("x_doubleSpinBox")
        self.gridLayout.addWidget(self.x_doubleSpinBox, 2, 4, 1, 2)
        self.v0_doubleSpinBox = QtWidgets.QDoubleSpinBox(parent=self.control_groupBox)
        self.v0_doubleSpinBox.setObjectName("v0_doubleSpinBox")
        self.gridLayout.addWidget(self.v0_doubleSpinBox, 6, 2, 1, 1)
        self.h_axis_comboBox = QtWidgets.QComboBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.h_axis_comboBox.sizePolicy().hasHeightForWidth()
        )
        self.h_axis_comboBox.setSizePolicy(sizePolicy)
        self.h_axis_comboBox.setObjectName("h_axis_comboBox")
        self.gridLayout.addWidget(self.h_axis_comboBox, 4, 1, 1, 1)
        self.label_3 = QtWidgets.QLabel(parent=self.control_groupBox)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 6, 6, 1, 1)
        self.scan_type_comboBox = QtWidgets.QComboBox(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.scan_type_comboBox.sizePolicy().hasHeightForWidth()
        )
        self.scan_type_comboBox.setSizePolicy(sizePolicy)
        self.scan_type_comboBox.setObjectName("scan_type_comboBox")
        self.gridLayout.addWidget(self.scan_type_comboBox, 0, 0, 1, 4)
        self.label = QtWidgets.QLabel(parent=self.control_groupBox)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 0, 1, 3)
        self.settings_horizontalLayout.addWidget(self.control_groupBox)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Minimum
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setMinimumSize(QtCore.QSize(100, 0))
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.continuous_scan_checkBox = QtWidgets.QCheckBox(parent=self.groupBox)
        self.continuous_scan_checkBox.setObjectName("continuous_scan_checkBox")
        self.gridLayout_3.addWidget(self.continuous_scan_checkBox, 0, 0, 1, 1)
        self.save_h5_checkBox = QtWidgets.QCheckBox(parent=self.groupBox)
        self.save_h5_checkBox.setObjectName("save_h5_checkBox")
        self.gridLayout_3.addWidget(self.save_h5_checkBox, 1, 0, 1, 1)
        self.verticalLayout_3.addWidget(self.groupBox)
        self.groupBox_2 = QtWidgets.QGroupBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setMinimumSize(QtCore.QSize(100, 0))
        self.groupBox_2.setObjectName("groupBox_2")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.groupBox_2)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.show_previous_scans_checkBox = QtWidgets.QCheckBox(parent=self.groupBox_2)
        self.show_previous_scans_checkBox.setObjectName("show_previous_scans_checkBox")
        self.gridLayout_4.addWidget(self.show_previous_scans_checkBox, 0, 0, 1, 1)
        self.clear_previous_scans_pushButton = QtWidgets.QPushButton(
            parent=self.groupBox_2
        )
        self.clear_previous_scans_pushButton.setObjectName(
            "clear_previous_scans_pushButton"
        )
        self.gridLayout_4.addWidget(self.clear_previous_scans_pushButton, 1, 0, 1, 1)
        self.verticalLayout_3.addWidget(self.groupBox_2)
        self.settings_horizontalLayout.addLayout(self.verticalLayout_3)
        self.details_groupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        self.details_groupBox.setObjectName("details_groupBox")
        self.details_layout = QtWidgets.QVBoxLayout(self.details_groupBox)
        self.details_layout.setObjectName("details_layout")
        self.settings_horizontalLayout.addWidget(self.details_groupBox)
        self.device_details_GroupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.Preferred,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.device_details_GroupBox.sizePolicy().hasHeightForWidth()
        )
        self.device_details_GroupBox.setSizePolicy(sizePolicy)
        self.device_details_GroupBox.setObjectName("device_details_GroupBox")
        self.device_details_layout = QtWidgets.QGridLayout(self.device_details_GroupBox)
        self.device_details_layout.setObjectName("device_details_layout")
        self.settings_horizontalLayout.addWidget(self.device_details_GroupBox)
        self.gridLayout_2.addLayout(self.settings_horizontalLayout, 0, 0, 1, 1)
        self.plot_groupBox = QtWidgets.QGroupBox(parent=self.centralwidget)
        self.plot_groupBox.setTitle("")
        self.plot_groupBox.setFlat(True)
        self.plot_groupBox.setObjectName("plot_groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.plot_groupBox)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout_2.addWidget(self.plot_groupBox, 1, 0, 1, 1)
        SimpleXYScanWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=SimpleXYScanWindow)
        self.menubar.setEnabled(True)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 692, 17))
        self.menubar.setObjectName("menubar")
        SimpleXYScanWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(parent=SimpleXYScanWindow)
        self.statusbar.setEnabled(True)
        self.statusbar.setObjectName("statusbar")
        SimpleXYScanWindow.setStatusBar(self.statusbar)

        self.retranslateUi(SimpleXYScanWindow)
        QtCore.QMetaObject.connectSlotsByName(SimpleXYScanWindow)
        SimpleXYScanWindow.setTabOrder(self.start_pushButton, self.h0_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.h0_doubleSpinBox, self.h1_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.h1_doubleSpinBox, self.dh_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.dh_doubleSpinBox, self.Nh_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.Nh_doubleSpinBox, self.v0_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.v0_doubleSpinBox, self.v1_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.v1_doubleSpinBox, self.dv_doubleSpinBox)
        SimpleXYScanWindow.setTabOrder(self.dv_doubleSpinBox, self.Nv_doubleSpinBox)

    def retranslateUi(self, SimpleXYScanWindow):
        _translate = QtCore.QCoreApplication.translate
        SimpleXYScanWindow.setWindowTitle(
            _translate("SimpleXYScanWindow", "Simple XY Scan")
        )
        self.control_groupBox.setTitle(_translate("SimpleXYScanWindow", "Scan Control"))
        self.x_label.setText(_translate("SimpleXYScanWindow", "current H"))
        self.dv_label_2.setText(_translate("SimpleXYScanWindow", "dv"))
        self.h_label.setText(_translate("SimpleXYScanWindow", "H"))
        self.v_label_2.setText(_translate("SimpleXYScanWindow", "V"))
        self.start_pushButton.setText(_translate("SimpleXYScanWindow", "Start"))
        self.dv_label.setText(_translate("SimpleXYScanWindow", "dh"))
        self.y_label.setText(_translate("SimpleXYScanWindow", "current V"))
        self.label_2.setText(_translate("SimpleXYScanWindow", "Nh"))
        self.label_3.setText(_translate("SimpleXYScanWindow", "Nv"))
        self.label.setText(_translate("SimpleXYScanWindow", "Progress"))
        self.groupBox.setTitle(_translate("SimpleXYScanWindow", "Data"))
        self.continuous_scan_checkBox.setText(
            _translate("SimpleXYScanWindow", "Continuous")
        )
        self.save_h5_checkBox.setText(_translate("SimpleXYScanWindow", "Save H5"))
        self.groupBox_2.setTitle(_translate("SimpleXYScanWindow", "View"))
        self.show_previous_scans_checkBox.setText(
            _translate("SimpleXYScanWindow", "Show \n" "Prev Scans")
        )
        self.clear_previous_scans_pushButton.setText(
            _translate("SimpleXYScanWindow", "Clear Prev")
        )
        self.details_groupBox.setTitle(_translate("SimpleXYScanWindow", "Details"))
        self.device_details_GroupBox.setTitle(
            _translate("SimpleXYScanWindow", "Device Details")
        )
